#include <stdio.h>
#include <assert.h>
#include "mess.h"
#include "btree.h"


/**********************************************************************/
/* This module provides an equilized binary tree for the messages,    */
/* sorted by priority. If you don't know how a binary tree works then */
/* don't mess with this file.                                         */
/**********************************************************************/

static BTree_Struct *mRoot=NULL;
static int16        MessCount=0;

void AddMess(Message_Struct* Mess)
{
    if (mRoot EQ NULL)
    {
        mRoot=(BTree_Struct*)MemAlloc(sizeof(BTree_Struct));
        mRoot->Right=NULL;
        mRoot->Left=NULL;
        mRoot->NodeStatus=0;
        mRoot->Up=NULL;
        mRoot->MessData=(MessStack*)MemAlloc(sizeof(MessStack));
        mRoot->MessData->Succ = NULL;
        mRoot->MessData->Mess = *Mess;
        mRoot->Priority       = Mess->Priority;
        mRoot->Type           = RIGHT;
    }
    else
    {
#ifndef NDEBUG
#if 0
        printf("Adding message \n%s\n%s\n%s\n",Mess->Lines[0], Mess->Lines[1], Mess->Lines[2]);
#endif
	CheckTree(mRoot);
#endif
        PushMess(Mess,mRoot);
    } 
}


void CheckTree(BTree_Struct* pNode)
{
    if (abs(pNode->NodeStatus)>1)
	printf("Tree is not equalized\n");
    if (pNode->MessData EQ NULL)
	printf("Node without message data detected\n");
    if (   ((pNode->NodeStatus < 0) AND (pNode->Left EQ NULL))
        OR ((pNode->NodeStatus > 0) AND (pNode->Right EQ NULL))
       )
	printf("non-NULL pointer expected\n");
    if (pNode->Left NEQ NULL)
	CheckTree(pNode->Left);
    if (pNode->Right NEQ NULL)
	CheckTree(pNode->Right);
}

void RotateRight(BTree_Struct* A)
{
    BTree_Struct **HelpNode1,
                 *B, *E;


    assert(A NEQ NULL);

    if (A->Up NEQ NULL)
        if (A->Type EQ LEFT)
            HelpNode1=&A->Up->Left;
        else
            HelpNode1=&A->Up->Right;
    else
        HelpNode1=&mRoot;

    if (A->Left->NodeStatus<=0)            /*       A            B       */
    {                                      /*      / \          / \      */
        *HelpNode1 = A->Left;              /*     B   C   ->   D   A     */
        B          = A->Left;              /*    / \              / \    */
        A->Left    = B->Right;             /*   D   E            E   C   */
        B->Right   = A;                    /*  n+1  n                    */
        B->Up      = A->Up;
        A->Up      = B;
        if (A->Left NEQ NULL)
        {
            A->Left->Up= A;
            A->Left->Type    = LEFT;
        }
        B->Type    = A->Type;
        A->Type    = RIGHT;

        (B->NodeStatus)++;
        A->NodeStatus=0-B->NodeStatus;
    }
    else                                     /*      A            E        */
    {                                        /*     / \         /   \      */
        B            = A->Left;              /*    B   C  ->   B     A     */
        E            = B->Right;             /*   / \         / \   / \    */
        A->Left      = E->Right;             /*  D   E       D   F G   C   */
        B->Right     = E->Left;              /*     / \                    */
        E->Left      = B;                    /*    F   G                   */
        E->Right     = A;
        *HelpNode1   = E;
        E->Up        = A->Up;
        
        if (B->Right NEQ NULL)
        {
            B->Right->Up = B;
            B->Right->Type=RIGHT;
        }
        if (A->Left NEQ NULL)
        {
            A->Left->Up  = A;
            A->Left->Type=LEFT;
        }
        B->Up=A->Up  = E;
        E->Type      = A->Type;
        A->Type      = RIGHT;
        
        if (E->NodeStatus>=0)
        {
            A->NodeStatus=0;
            B->NodeStatus=0-E->NodeStatus;
            E->NodeStatus=0;
        }
        else
        {
            A->NodeStatus=1;
            B->NodeStatus=0;
            E->NodeStatus=0;
        }
    }
}


void RotateLeft(BTree_Struct* A)
{
    BTree_Struct **HelpNode1,
                 *C,*D;


    assert(A NEQ NULL);

    if (A->Up NEQ NULL)
    if (A->Type EQ LEFT)
            HelpNode1=&A->Up->Left;
        else
            HelpNode1=&A->Up->Right;
    else
        HelpNode1=&mRoot;

    if (A->Right->NodeStatus>=0)             /*     A             C       */
    {                                        /*    / \           / \      */
        *HelpNode1   = A->Right;             /*   B   C   ->    A   E     */
        C            = A->Right;             /*      / \       / \        */
        A->Right     = C->Left;              /*     D   E     B   D       */
        C->Left      = A;                    /*     n  n+1                */
        C->Up        = A->Up;
        A->Up        = C;
        if (A->Right NEQ NULL)
        {
            A->Right->Up = A;
            A->Right->Type=RIGHT;
        }
        C->Type      = A->Type;
        A->Type      = LEFT;
        
        (C->NodeStatus)--;
        A->NodeStatus=0-C->NodeStatus;
    }
    else                                     /*    A              D       */
    {                                        /*   / \           /   \     */
        C            = A->Right;             /*  B   C    ->   A     C    */
        D            = C->Left;              /*     / \       / \   / \   */
        A->Right     = D->Left;              /*    D   E     B   F G   E  */
        C->Left      = D->Right;             /*   / \                     */
        D->Left      = A;                    /*  F   G                    */
        D->Right     = C;
        *HelpNode1   = D;
        D->Up        = A->Up;
        if (A->Right NEQ NULL)
        {
            A->Right->Up = A;
            A->Right->Type=RIGHT;
        }
        if (C->Left NEQ NULL)
        {
            C->Left->Up  = C;
            C->Left->Type  = LEFT;
        }
        A->Up=C->Up  = D;
        D->Type      = A->Type;
        A->Type      = LEFT;
        
        if (D->NodeStatus<=0)
        {
            A->NodeStatus=0;
            C->NodeStatus=0-D->NodeStatus;
            D->NodeStatus=0;
        }
        else
        {
            A->NodeStatus=-1;
            C->NodeStatus=0;
            D->NodeStatus=0;
        }
    }
}

void LiftLeft(BTree_Struct* Node)
{
    assert(Node NEQ NULL);

    (Node->NodeStatus)--;
    if (Node->NodeStatus EQ -2)
        RotateRight(Node);
    else
        if (Node->Up NEQ NULL)
            if (Node->Type EQ LEFT)
                LiftLeft(Node->Up);
            else
                LiftRight(Node->Up);
}

void LiftRight(BTree_Struct* Node)
{
    assert(Node NEQ NULL);

    (Node->NodeStatus)++;
    if (Node->NodeStatus EQ 2)
        RotateLeft(Node);
    else
        if (Node->Up NEQ NULL)
            if (Node->Type EQ LEFT)
                LiftLeft(Node->Up);
            else
                LiftRight(Node->Up);
}


void PushMess(Message_Struct* pMess, BTree_Struct* Node)
{ 
    MessStack  *MessSt;

    assert((Node NEQ NULL) AND (pMess NEQ NULL));

    if (Node->Priority EQ pMess->Priority)
    {
        MessSt=Node->MessData;
        while (MessSt->Succ NEQ NULL)
            MessSt=MessSt->Succ;

        MessSt->Succ   = (MessStack*)MemAlloc(sizeof(MessStack));
        MessSt         = MessSt->Succ;
        MessSt->Succ   = NULL;
        MessSt->Mess   = *pMess;
    }
    else
    {
        if (Node->Priority > pMess->Priority)
        {
            if (Node->Left NEQ NULL)
                PushMess(pMess,Node->Left);
            else
            {
                Node->NodeStatus--;
                Node->Left                  = (BTree_Struct*)MemAlloc(sizeof(BTree_Struct));
                Node->Left->Up              = Node;
                Node->Left->Left            = NULL;
                Node->Left->Right           = NULL;
                Node->Left->Priority        = pMess->Priority;
                Node->Left->MessData        = (MessStack*)MemAlloc(sizeof(MessStack));
                Node->Left->MessData->Mess  = *pMess;
                Node->Left->MessData->Succ  = NULL;
                Node->Left->Type            = LEFT;
                Node->Left->NodeStatus      = 0;

                if ((Node->Right EQ NULL) AND (Node->Up NEQ NULL))
                {
                    if (Node->Type EQ LEFT)
                        LiftLeft(Node->Up);
                    else
                        LiftRight(Node->Up);
                }
            }
        }
        else
        {
            if (Node->Right NEQ NULL)
                PushMess(pMess,Node->Right);
            else
            {
                Node->Right                  = (BTree_Struct*)MemAlloc(sizeof(BTree_Struct));
                Node->Right->Up              = Node;
                Node->Right->Left            = NULL;
                Node->Right->Right           = NULL;
                Node->Right->Priority        = pMess->Priority;
                Node->Right->MessData        = (MessStack*)MemAlloc(sizeof(MessStack));
                Node->Right->MessData->Mess  = *pMess;
                Node->Right->MessData->Succ  = NULL;
                Node->Right->Type            = RIGHT;
                Node->Right->NodeStatus      = 0;

                Node->NodeStatus++;
                if ((Node->Left EQ NULL) AND (Node->Up NEQ NULL))
                {
                    if (Node->Type EQ LEFT)
                        LiftLeft(Node->Up);
                    else
                        LiftRight(Node->Up);
                }
            }

        }
    }
}


void ClearNode(BTree_Struct* Node)
{
    static MessStack *StackPtr;

    assert(Node NEQ NULL);

    if (Node->Left NEQ NULL)
    {
        ClearNode(Node->Left);
        MemFree(Node->Left);
    }

    StackPtr=Node->MessData;
    while (StackPtr NEQ NULL)
    {
        gMessages[MessCount]=Node->MessData->Mess;
        MessCount++;
        StackPtr = StackPtr->Succ;
        MemFree(Node->MessData);
        Node->MessData = StackPtr;
    }

    if (Node->Right NEQ NULL)
    {
        ClearNode(Node->Right);
        MemFree(Node->Right);
    }
}

void ClearMessTree()
{
    if (mRoot NEQ NULL)
    {
        ClearNode(mRoot);
        MemFree(mRoot);
    }
    
    gMessageNr=MessCount;
    MessCount=0;
    mRoot=NULL;
}




