#ifndef _MESS_H_
#define _MESS_H_

#include "stdtyp.h"

#ifndef MAXDIR
#define MAXDIR 255
#endif

#ifndef MAXNAMELEN
#define MAXNAMELEN 255
#endif

#ifndef max
#define max(a,b)    (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif

#ifndef abs
#define abs(a)      (((a) > 0) ? (a) : (-(a)))
#endif

#undef NULL
#define NULL (void*)0

#define IsDigit(c) (isdigit(c) OR ((c) EQ '-') OR ((c) EQ '+') )

typedef enum { PASS=1, SUMMARY=2, DOUBLE=3, DISCARD=4 }
MessageType;

typedef enum { IO_FAILURE, IO_SUCCESS }
IO_Def;


typedef struct
{
    int16  Length,
           LineNr,
           Priority;
    char   **Lines;
}
Message_Struct;


typedef struct
{
    int16        Line,
                 Word;
    char         *Text;
}
DefLine_Struct;

typedef struct
{
    int16           Type,
                    IdentNr,
                    NonIdentNr,
                    IgnoreNr,
                    SummaryNr,
                    HeaderNr,
                    CompareNr,
                    Priority;
    DefLine_Struct  *Ident;
    DefLine_Struct  *NonIdent;
    DefLine_Struct  *Ignore;
    DefLine_Struct  *Summary;
    int16           *Compare;
    char            **SummaryHeader;
    char            *SummaryFormat;
}
MessDef_Struct;

typedef struct MessDef_Stack
{
           MessDef_Struct  Data;
    struct MessDef_Stack   *Succ;
} MessDef_Stack;


extern const char  *MessDefStrings[4];
extern const char  *IgnoreStr;
extern const char  *IdentStr;
extern const char  *NIdentStr;
extern const char  *SummaryStr;
extern const char  *HeaderStr;
extern const char  *SummaryLineStr;
extern const char  *PriorityStr;
extern const char  *CompareStr;
extern const int16 MaxLineNr,
                   MaxRowNr;

extern char        MessFile[20];
extern char        HostPath[100];
extern const char  gVersionStr[10];


extern Message_Struct  *gMessages;
extern MessDef_Struct  *gMessDefs;

extern int16           gMessageNr,
                       gAuxMessNr,
                       DefCount;


extern char mResultName[20];

/* global functions */

/* messio.c : */
IO_Def Read_Messages(int16 player);
IO_Def Write_Messages(int16 player);
IO_Def ReadMessDefs(void);

/* btree.c */
void AddMess(Message_Struct* Mess);
void ClearMessTree(void);

/* util.h */
void* MemAlloc(size_t n);
void* MemCAlloc(size_t num, size_t size);
void  MemFree(void* ptr);
char* STRcpy(char* s1, char* s2);
#endif
