#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "conf.h"

#ifdef NEED_DIR_H
#include <dir.h>
#endif

#ifdef NEED_UNISTD_H
#include <unistd.h>
#endif

#include "btree.h"
#include "mess.h"
#include "util.h"

const char  *MessDefStrings[4] = {"[PASS]","[SUMMARY]","[DOUBLE]","[DISCARD]"};
const char  *IgnoreStr         = "[IGNORE]";
const char  *IdentStr          = "[ID]";
const char  *NIdentStr         = "[IDN]";
const char  *SummaryStr        = "[EXTRACT]";
const char  *HeaderStr         = "[HEADER]";
const char  *SummaryLineStr    = "[LINE]";
const char  *PriorityStr       = "[PRIORITY]";
const char  *CompareStr        = "[COMPARE]";
const int16 MaxLineNr          = 20;
const int16 MaxRowNr           = 40;

char            MessFile[20]   = "mess.def";
const char      gVersionStr[10]= "1.4beta";
char            *AuxMessFile   = NULL;
MessStack       *gAuxMess      = NULL;


Message_Struct  *gMessages;
MessDef_Struct  *gMessDefs;

int16           gMessageNr,
                gAuxMessNr=0,
                DefCount=0;


/*************************************************************************/
/*                        Miscellanous stuff                             */
/*************************************************************************/


/* free message struct, used when a message is deleted */
void FreeMessage(Message_Struct Mess)
{
    int16 i;

    for (i=0; i<Mess.LineNr; i++)
    {
        assert(Mess.Lines[i] != NULL);
        MemFree(Mess.Lines[i]);
    }
    assert(Mess.Lines != NULL);
    MemFree(Mess.Lines);
}


/* delete a specific word in a text line */
char* DelWord(char* Buffer, int16 WordNr)
{
    char   *s1=(char*)MemAlloc(1001),
           *s2;
    int16  i;


    if (WordNr<0)
    {
        strcpy(s1,Buffer);
        WordNr+=2;
        strtok(s1," ");

        while (strtok(NULL," ") NEQ NULL)
            WordNr++;
    }

    if (WordNr<=0)
        goto EndDelWord;

    strcpy(s1,Buffer);

    if (WordNr EQ 1)
    {
        strtok(s1," ");
        if ((s2=strtok(NULL," ")) EQ NULL)
            Buffer[0]=0;
        else
            strcpy(Buffer,s2);
    }
    else
    {
        strcpy(Buffer,strtok(s1," "));
        for (i=0; ((i<(WordNr-1)) AND (s2=strtok(NULL," ")) NEQ NULL); i++)
            strcat(Buffer,s2);

        strtok(NULL," ");
        while ((s2=strtok(NULL," ")) NEQ NULL)
            strcat(Buffer,s2);
    }

EndDelWord:
    MemFree(s1);
    return(Buffer);
}


/* Function Compare_Messages :
 * This function will take a definition, consider [COMPARE] and [IGNORE]
 * commands and return whether the messages match
 *
 * Example : If we have Dark Sense messages with the standard mess.def
 *           file then 2 Dark Senses are considered equal if and only if
 *           they scan the same planet.
 */
boolean Compare_Messages(Message_Struct *Message1, Message_Struct *Message2,
                         MessDef_Struct *pMessDef)
{
    int16           i,j;
    char            *MessBuffer=NULL, 
                    *BufPos=NULL;
    Message_Struct  Mess1, Mess2;


    Mess1.LineNr=Message1->LineNr;
    Mess2.LineNr=Message2->LineNr;
    Mess1.Lines=(char**)MemAlloc(Mess1.LineNr*sizeof(char*));
    Mess2.Lines=(char**)MemAlloc(Mess2.LineNr*sizeof(char*));

    if (pMessDef->CompareNr>0)
    {
        memset(Mess1.Lines,0,Mess1.LineNr*sizeof(char*));
        memset(Mess2.Lines,0,Mess2.LineNr*sizeof(char*));

        for (j=0; j<pMessDef->CompareNr; j++)
        {
            Mess1.Lines[pMessDef->Compare[j]]=Message1->Lines[pMessDef->Compare[j]];
            Mess2.Lines[pMessDef->Compare[j]]=Message2->Lines[pMessDef->Compare[j]];
        }
    }
    else
    {
        memcpy(Mess1.Lines,Message1->Lines,Mess1.LineNr*sizeof(char*));
        memcpy(Mess2.Lines,Message2->Lines,Mess2.LineNr*sizeof(char*));
    }

    for (j=0; j<pMessDef->IgnoreNr; j++)
      if ((Mess1.Lines[pMessDef->Ignore[j].Line] NEQ NULL) OR (Mess2.Lines[pMessDef->Ignore[j].Line] NEQ NULL))
        if (pMessDef->Ignore[j].Word EQ 0)
        {
            if (Mess1.LineNr>pMessDef->Ignore[j].Line)
                Mess1.Lines[pMessDef->Ignore[j].Line]=NULL;

            if (Mess2.LineNr>=pMessDef->Ignore[j].Line)
                Mess2.Lines[pMessDef->Ignore[j].Line]=NULL;
        }
        else
        {
            if ((Mess1.LineNr>pMessDef->Ignore[j].Line) AND (Mess1.Lines[pMessDef->Ignore[j].Line] NEQ NULL))
            {
                if (MessBuffer EQ NULL)
                {
                    MessBuffer=(char*)MemAlloc(1001);
                    BufPos=MessBuffer;
                }

                strcpy(BufPos,Mess1.Lines[pMessDef->Ignore[j].Line]);
                Mess1.Lines[pMessDef->Ignore[j].Line]=BufPos;
                DelWord(BufPos,pMessDef->Ignore[j].Word);
                BufPos+=strlen(BufPos)+1;
            }

            if ((Mess2.LineNr>pMessDef->Ignore[j].Line) AND (Mess2.Lines[pMessDef->Ignore[j].Line] NEQ NULL))
            {
                if (MessBuffer EQ NULL)
                {
                    MessBuffer=(char*)MemAlloc(1001);
                    BufPos=MessBuffer;
                }

                strcpy(BufPos,Mess2.Lines[pMessDef->Ignore[j].Line]);
                Mess2.Lines[pMessDef->Ignore[j].Line]=BufPos;
                DelWord(BufPos,pMessDef->Ignore[j].Word);
                BufPos+=strlen(BufPos)+1;
            }
        }


    for (i=0; i<max(Mess1.LineNr,Mess2.LineNr); i++)
        if ((Mess1.Lines[i] NEQ NULL) OR (Mess2.Lines[i] NEQ NULL))
        {
            if (     (Mess1.Lines[i] EQ NULL)
                 AND (strspn(Mess2.Lines[i]," ") NEQ strlen(Mess2.Lines[i])) )
                goto False;

            if (     (Mess2.Lines[i] EQ NULL)
                 AND (strspn(Mess1.Lines[i]," ") NEQ strlen(Mess1.Lines[i])) )
                goto False;

            if (     (Mess1.Lines[i] NEQ NULL) AND (Mess2.Lines[i] NEQ NULL)
                 AND strcmp(Mess1.Lines[i],Mess2.Lines[i]) )
                goto False;

        }

    if (MessBuffer NEQ NULL)
        MemFree(MessBuffer);
    MemFree(Mess1.Lines);
    MemFree(Mess2.Lines);

    return(true);

False:
    if (MessBuffer NEQ NULL)
        MemFree(MessBuffer);
    MemFree(Mess1.Lines);
    MemFree(Mess2.Lines);
    return(false);
}





/* Function Identify_Message :
 * This function takes a message and a message definition and returns if
 * the definition can be applied to the message or not
 */
boolean Identify_Message(Message_Struct Mess, int16 ID_NR, DefLine_Struct *Ident,
                         int16 IDN_NR, DefLine_Struct *NIdent)
{
    char      *Buffer=(char*)MemAlloc(1001);
              char      *s1=NULL;
              int16     i, j, WordCount;

    for (i=0; i<ID_NR; i++)
        if (Mess.LineNr>Ident[i].Line)
        {
            strcpy(Buffer,Mess.Lines[Ident[i].Line]);
            if (Ident[i].Word EQ 0)
            {
                j=0;
                while (Buffer[j] NEQ 0)
                {
                    Buffer[j]=toupper(Buffer[j]);
                    j++;
                }
                if (strstr(Buffer,Ident[i].Text) EQ NULL)
                    goto EndIdentify;
            }
            else
                if ((WordCount=Ident[i].Word)<0)
                {
		    s1=(char*)MemAlloc(1001);
                    strcpy(s1,Buffer);
                    WordCount=2;
                    strtok(s1," ");
                    while (strtok(NULL," ") NEQ NULL)
                        WordCount++;
		    free(s1);

                    WordCount+=Ident[i].Word;
                    if (WordCount>0)
                    {
                        /* copy wanted word to s1 */
                        if (WordCount EQ 1)
                            s1=strtok(Buffer," ");
                        else
                        {
                            strtok(Buffer," ");
                            for (j=1; j<WordCount; j++)
                                s1=strtok(NULL," ");
                        }

                        j=0;
                        while (s1[j] NEQ 0)
                        {
                            s1[j]=toupper(s1[j]);
                            j++;
                        }

                        if (strcmp(s1,Ident[i].Text) NEQ 0)
                            goto EndIdentify;
                    }
                }
                else
                {
                    /* copy wanted word to s1 */
                    if (WordCount EQ 1)
                        s1=strtok(Buffer," ");
                    else
                    {
                        strtok(Buffer," ");
                        for (j=1; j<WordCount; j++)
                            s1=strtok(NULL," ");
                    }

                    j=0;
                    while (s1[j] NEQ 0)
                    {
                        s1[j]=toupper(s1[j]);
                        j++;
                    }

                    if (strcmp(s1,Ident[i].Text) NEQ 0)
                        goto EndIdentify;
                }
        }
        else
            goto EndIdentify;


/* Non-Ident */


    for (i=0; i<IDN_NR; i++)
        if (Mess.LineNr>NIdent[i].Line)
        {
            strcpy(Buffer,Mess.Lines[NIdent[i].Line]);
            if (NIdent[i].Word EQ 0)
            {
                j=0;
                while (Buffer[j] NEQ 0)
                {
                    Buffer[j]=toupper(Buffer[j]);
                    j++;
                }

                if (strstr(Buffer,NIdent[i].Text) NEQ NULL)
                    goto EndIdentify;
            }
            else
                if ((WordCount=NIdent[i].Word)<0)
                {
		    s1=(char*)MemAlloc(1001);
                    strcpy(s1,Buffer);
                    WordCount=2;
                    strtok(s1," ");
                    while (strtok(NULL," ") NEQ NULL)
                        WordCount++;
		    free(s1);

                    WordCount+=NIdent[i].Word;

                    if (WordCount>0)
                    {
                        /* copy wanted word to s1 */
                        if (WordCount EQ 1)
                            s1=strtok(Buffer," ");
                        else
                        {
                            strtok(Buffer," ");
                            for (j=1; j<WordCount; j++)
                                s1=strtok(NULL," ");
                        }

                        j=0;
                        while (s1[j] NEQ 0)
                            s1[j]=toupper(s1[j++]);

                        if (strcmp(s1,NIdent[i].Text) EQ 0)
                            goto EndIdentify;
                    }
                }
                else
                {
                    /* copy wanted word to s1 */
                    if (WordCount EQ 1)
                        s1=strtok(Buffer," ");
                    else
                    {
                        strtok(Buffer," ");
                        for (j=1; j<WordCount; j++)
                            s1=strtok(NULL," ");
                    }

                    j=0;
                    while (s1[j] NEQ 0)
                        s1[j]=toupper(s1[j++]);

                    if (strcmp(s1,NIdent[i].Text) EQ 0)
                        goto EndIdentify;
                }
        }
    MemFree(Buffer);
    return(true);

EndIdentify:
    MemFree(Buffer);
    return false;

}




/* The function Filter_Double will take a certain definition of type [DOUBLE]
 * and eliminates all double messages that match this definition.
 */
void Filter_Double(int16 DefNr)
{
    Message_Struct  **NewMess;
    int16           DoubleCount=0,
                    i,j;

    NewMess=(Message_Struct**)MemAlloc(gMessageNr*sizeof(Message_Struct*));

    for (i=gMessageNr-1; i>=0; i--)
        if (gMessages[i].Length NEQ 0)
            if (Identify_Message(gMessages[i], gMessDefs[DefNr].IdentNr, gMessDefs[DefNr].Ident,
                                 gMessDefs[DefNr].NonIdentNr,gMessDefs[DefNr].NonIdent))
            {
                for (j=0; ((j<DoubleCount) AND (gMessages[i].Length NEQ 0)); j++)

                    if (Compare_Messages(NewMess[j],&gMessages[i],&gMessDefs[DefNr]))
                    {
                        gMessages[i].Length=0;
                        FreeMessage(gMessages[i]);
                    }

                if (gMessages[i].Length NEQ 0)
                {
                    gMessages[i].Priority=gMessDefs[DefNr].Priority;
                    NewMess[DoubleCount]=(Message_Struct*)MemAlloc(sizeof(Message_Struct));
                    *(NewMess[DoubleCount])=gMessages[i];
                    gMessages[i].Length=0;
                    DoubleCount++;
                }
            }
    for (i=0; i<DoubleCount; i++)
    {
        AddMess(NewMess[i]);
        MemFree(NewMess[i]);
    }

    MemFree(NewMess);
}



/* The function Filter_Summary will take a certain definition of type
 * [SUMMARY] and summarize all messages that match this definition.
 */
void Filter_Summary(int16 DefNr)
{
    Message_Struct  **NewMess;
    int16           i,j,k,
                    NewMessCount=0,
                    NewMessLine=MaxLineNr-1,
                    MessLine,
                    WordNr,
                    LinePos,
                    NewLinePos;
    char            *NewLine=(char*)MemAlloc(1001),
                    *Buffer=(char*)MemAlloc(1001),
                    *FormatStr=(char*)MemAlloc(1001),
                    *s1;

    NewMess=(Message_Struct**)MemAlloc(gMessageNr*sizeof(Message_Struct*));

    for (i=0; i<gMessageNr; i++)
        if (gMessages[i].Length NEQ 0)
            if (Identify_Message(gMessages[i], gMessDefs[DefNr].IdentNr, gMessDefs[DefNr].Ident,
                                 gMessDefs[DefNr].NonIdentNr,gMessDefs[DefNr].NonIdent))
            {
/* We can only summarize 18 or so messages into one message because there is
 * a limit to the number of lines that a message can have. If we reached
 * this count then we need a new summary message.
 */
                if (NewMessLine EQ MaxLineNr-1)
                {
                    NewMessCount++;
                    NewMess[NewMessCount-1]=(Message_Struct*)MemAlloc(sizeof(Message_Struct));
                    NewMess[NewMessCount-1]->Length=1;
                    NewMess[NewMessCount-1]->Priority=gMessDefs[DefNr].Priority;
                    NewMess[NewMessCount-1]->LineNr=gMessDefs[DefNr].HeaderNr;
                    NewMess[NewMessCount-1]->Lines=(char**)MemAlloc(MaxLineNr*sizeof(char*));

                    for (j=0; j<gMessDefs[DefNr].HeaderNr; j++)
                    {
                        NewMess[NewMessCount-1]->Lines[j]=(char*)MemAlloc(strlen(gMessDefs[DefNr].SummaryHeader[j])+1);
                        strcpy(NewMess[NewMessCount-1]->Lines[j],gMessDefs[DefNr].SummaryHeader[j]);
                    }
                    NewMessLine=j;
                }


/* Now get the items to summarize out of the message and generate our new
 * summary line. This is some work but goes straightforward.
 */

                LinePos=0;
                NewLinePos=0;
                for (j=0; j<gMessDefs[DefNr].SummaryNr; j++)
                {
                    strcpy(Buffer,gMessages[i].Lines[(MessLine=gMessDefs[DefNr].Summary[j].Line)]);
                    s1=NULL;
                    if (gMessages[i].LineNr>MessLine)
                        if (gMessDefs[DefNr].Summary[j].Word NEQ 0)
                        {
                            if ((WordNr=gMessDefs[DefNr].Summary[j].Word)<0)
                            {
                                s1=strtok(Buffer," ");
                                WordNr+=2;
                                while ((s1=strtok(NULL," ")) NEQ NULL)
                                    WordNr++;
                                strcpy(Buffer,gMessages[i].Lines[MessLine]);
                            }
                            if (WordNr NEQ 0)
                            {
                                s1=strtok(Buffer," ");
                                for (k=1; k<WordNr; k++)
                                    s1=strtok(NULL," ");
                            }
                        }
                        else
                            s1=Buffer;

                    while ((gMessDefs[DefNr].SummaryFormat[LinePos] NEQ '%') AND (gMessDefs[DefNr].SummaryFormat[LinePos] NEQ 0))
                    {
                        NewLine[NewLinePos]=gMessDefs[DefNr].SummaryFormat[LinePos];
                        LinePos++;
                        NewLinePos++;
                    }
                    LinePos++;
                    if (IsDigit(gMessDefs[DefNr].SummaryFormat[LinePos]))
                    {
                        k=atoi(&gMessDefs[DefNr].SummaryFormat[LinePos]);
                        LinePos++;
                        while (IsDigit(gMessDefs[DefNr].SummaryFormat[LinePos]))
                            LinePos++;
                    }
                    else
                        k=((s1 NEQ NULL) ? strlen(s1) : 0);

		    if (s1 NEQ NULL)
		    {
                        if (NewLinePos+abs(k)>40)
                            k=40-NewLinePos;
                        s1[abs(k)]=0;
                        sprintf(FormatStr,"%%%ds",k);
                        sprintf(&NewLine[NewLinePos],FormatStr,s1);
                        NewLinePos+=abs(k);
		    }

                }
/* Concat rest of format string */
                while ((gMessDefs[DefNr].SummaryFormat[LinePos] NEQ 0) AND (NewLinePos<40))
                {
                    NewLine[NewLinePos]=gMessDefs[DefNr].SummaryFormat[LinePos];
                    LinePos++;
                    NewLinePos++;
                }
/* Compare new Line to existing summary lines */

                for (k=0; k<NewMessCount; k++)
                    for (j=gMessDefs[DefNr].HeaderNr; ((j<NewMess[k]->LineNr) AND (NewLine[0])); j++)
                        if (!strcmp(NewLine,NewMess[k]->Lines[j]))
                            NewLine[0]=0;

                if (NewLine[0])
                {
                    (NewMess[NewMessCount-1]->LineNr)++;
                    NewLine[NewLinePos]=0;
                    NewMess[NewMessCount-1]->Lines[NewMessLine]=(char*)MemAlloc(NewLinePos+1);
                    strcpy(NewMess[NewMessCount-1]->Lines[NewMessLine],NewLine);
                    NewMessLine++;
               }
               gMessages[i].Length=0;
               FreeMessage(gMessages[i]);
           }

    for (i=0; i<NewMessCount; i++)
    {
        AddMess(NewMess[i]);
        MemFree(NewMess[i]);
    }

    MemFree(NewMess);
    MemFree(NewLine);
    MemFree(Buffer);
    MemFree(FormatStr);
}



/* The Filter() function goes through all message definition and applies
 * them to all messages. Messages that were copied are moved over to
 * the message tree, deleted messages get removed from the message array.
 * So a message that matches any definition won't be processed by any other
 * definition because it is either deleted, summarized (and the summary
 * message appears among the new messages) or copied.
 */
void Filter(void)
{
    int16           i,j;


    for (i=0; i<DefCount; i++)
    {
        switch(gMessDefs[i].Type)
        {
        case PASS    : for (j=0; j<gMessageNr; j++)
                           if (gMessages[j].Length NEQ 0)
                               if (Identify_Message(gMessages[j],gMessDefs[i].IdentNr,gMessDefs[i].Ident,
                                                    gMessDefs[i].NonIdentNr,gMessDefs[i].NonIdent))
                               {
                                   gMessages[j].Priority=gMessDefs[i].Priority;
                                   AddMess(&gMessages[j]);
                                   gMessages[j].Length=0;
                               }
                       break;
        case DISCARD : for (j=0; j<gMessageNr; j++)
                           if (gMessages[j].Length NEQ 0)
                               if (Identify_Message(gMessages[j],gMessDefs[i].IdentNr,gMessDefs[i].Ident,
                                                    gMessDefs[i].NonIdentNr,gMessDefs[i].NonIdent))
                               {
                                   gMessages[j].Length=0;
                                   FreeMessage(gMessages[j]);
                               }
                       break;
        case DOUBLE  : Filter_Double(i);
                       break;
        case SUMMARY : Filter_Summary(i);
                       break;
        }
    }

    for (j=0, i=0; i<gMessageNr; i++)
        if (gMessages[i].Length NEQ 0)
            AddMess(&gMessages[i]);

}

void ClearData(void)
{
    int16 i;

    for (i=0; i<gMessageNr; i++)
        FreeMessage(gMessages[i]);
    gMessageNr=0;

    MemFree(gMessages);
}



/* ReadAuxMess() will read an external message and break it into the planets
 * message format if necessary.
 */
void ReadAuxMess(void)
{
    FILE           *InpFile;
    char           lBuffer[MAXNAMELEN],
                   *lPtr,
                   *lMessLine;
    MessStack      *lStackPtr=NULL;

    assert(AuxMessFile NEQ NULL);
    if ((InpFile=fopen(AuxMessFile,"rt")) EQ NULL)
        return;

    while (fgets(lBuffer,MAXNAMELEN,InpFile))
    {
/* initialize structures */
        if (gAuxMess EQ NULL)
        {
            gAuxMess=(MessStack*)MemAlloc(sizeof(MessStack));
            memset(gAuxMess,0,sizeof(MessStack));
            gAuxMess->Mess.Lines=(char**)MemAlloc(MaxLineNr*sizeof(char*));
            lStackPtr=gAuxMess;
            gAuxMessNr++;
        }

        if (strlen(lBuffer)<MaxRowNr)
        {
            if (lStackPtr->Mess.LineNr EQ MaxLineNr)
            {
                lStackPtr->Succ=(MessStack*)MemAlloc(sizeof(MessStack));
                lStackPtr=lStackPtr->Succ;
                memset(lStackPtr,0,sizeof(MessStack));
                lStackPtr->Mess.Lines=(char**)MemAlloc(MaxLineNr*sizeof(char*));
                gAuxMessNr++;
            }
            lStackPtr->Mess.Lines[lStackPtr->Mess.LineNr]
                  = (lMessLine=(char*)MemAlloc((MaxRowNr+1)*sizeof(char)));
            strcpy(lMessLine,lBuffer);
            lStackPtr->Mess.LineNr++;
        }
        else
        {
            lPtr=strtok(lBuffer," \t");

    /* now fill as many lines as we need */
            while (lPtr NEQ NULL)
            {
                if (lStackPtr->Mess.LineNr EQ MaxLineNr)
                {
                    lStackPtr->Succ=(MessStack*)MemAlloc(sizeof(MessStack));
                    lStackPtr=lStackPtr->Succ;
                    memset(lStackPtr,0,sizeof(MessStack));
                    lStackPtr->Mess.Lines=(char**)MemAlloc(MaxLineNr*sizeof(char*));
                    gAuxMessNr++;
                }

    /* processing of new line starts */
                lStackPtr->Mess.Lines[lStackPtr->Mess.LineNr]
                      = (lMessLine=(char*)MemAlloc((MaxRowNr+1)*sizeof(char)));
                lMessLine[0]=0;

                while (    (lPtr NEQ NULL)
                       AND (   (strlen(lMessLine) EQ 0)
                            OR (strlen(lMessLine) + strlen(lPtr) < MaxRowNr)
                      )    )
                      
                {
                    if (strlen(lMessLine))    /* insert space */
                        strcat(lMessLine," ");
                    strncat(lMessLine,lPtr,MaxRowNr-strlen(lMessLine));
                    lPtr=strtok(NULL," \t");
                }
                lStackPtr->Mess.LineNr++;
            }
        }
    }
}



/* The following function adds all external messages to the message tree.
 */
void SendMess(void)
{
    MessStack         *lStackPtr=gAuxMess;
    Message_Struct    *lMess=(Message_Struct*)MemAlloc(sizeof(Message_Struct));
    int16             lCount;

    while (lStackPtr NEQ NULL)
    {
        lStackPtr->Mess.Priority=-1;
        memcpy(lMess,&lStackPtr->Mess,sizeof(Message_Struct));
        lMess->Lines=(char**)MemAlloc(sizeof(char*)*lStackPtr->Mess.LineNr);
        for (lCount=0; lCount<lStackPtr->Mess.LineNr; lCount++)
        {
            lMess->Lines[lCount]=(char*)MemAlloc(strlen(lStackPtr->Mess.Lines[lCount])+1);
            strcpy(lMess->Lines[lCount],lStackPtr->Mess.Lines[lCount]);
	    }
        AddMess(lMess);
        lStackPtr=lStackPtr->Succ;
    }
}



void ClearAuxMess(void)
{
    MessStack *lMessPtr=gAuxMess;

    while (lMessPtr NEQ NULL)
    {
        gAuxMess=lMessPtr->Succ;
        MemFree(lMessPtr);
        lMessPtr=gAuxMess;
    }
}




int main(int16 argc, char** argv)
{
    int16            i,j,
                     lPlayersToFilter[11]={1,1,1,1,1,1,1,1,1,1,1};
    char             lPathToResults[MAXDIR],
                     lPreviousPath[MAXDIR];
    boolean          lSilentOperation=false;

    static const char playerdef[11]="123456789ab";

    strcpy(lPathToResults,".");
    for (i=argc-1; i>0; i--)
    {
        if (argv[i][0] NEQ '-')
            strncpy(lPathToResults,argv[i],MAXDIR);
        else
            switch(toupper(argv[i][1]))
            {
            case 'P' : memset(lPlayersToFilter,0,sizeof(int16)*11);
                       for (j=1; j<=11; j++)
                           lPlayersToFilter[j-1]=(strchr(argv[i],playerdef[j-1])!=NULL);
                       break;

            case 'D' : strcpy(MessFile,&argv[i][2]);
                       break;

            case 'S' : if (AuxMessFile==NULL)
                           AuxMessFile=(char*)MemAlloc(strlen(&argv[i][2])+1);
                       else
                           AuxMessFile=(char*)realloc(AuxMessFile,strlen(&argv[i][2])+1);
                       strcpy(AuxMessFile,&argv[i][2]);
                       break;

	    case 'Q' : lSilentOperation=true;
		       break;

            case 'V' : printf("\n\nThis is MESS %s, compiled on " __DATE__ " at " __TIME__ "\n\n",gVersionStr);
                       exit(0);

            case 'H' : printf("\n\nThis is MESS %s, compiled on " __DATE__ " at " __TIME__ "\n\n",gVersionStr);
                       printf("Usage : MESS [options] [PathToResults]\n");
                       printf("Valid options :\n"
                              "      -p<Playerlist> : Filter only specified results (e.g. -p1347a)\n"
                              "      -d<DefFile>    : use <DefFile> instead of MESS.DEF (e.g. -dnew.def)\n"
                              "      -s<Textfile>   : insert <Textfile> as player message\n"
			      "      -q             : quit operation\n"
                              "      -h             : this help\n"
                              "      -v             : print version number and exit\n");
                       exit(0);

            }
    }


    getcwd(lPreviousPath,MAXDIR);
    if (!lSilentOperation)
        printf("\n\nreading message definition file : ");

    if (ReadMessDefs() EQ IO_SUCCESS)
    {
	if (!lSilentOperation)
            printf("%d definitions read\n",DefCount);
        if (AuxMessFile NEQ NULL)
            ReadAuxMess();

        chdir(lPathToResults);

        for (i=1; i<=11; i++)
            if (lPlayersToFilter[i-1])
                if (Read_Messages(i) EQ IO_SUCCESS)
                {
		    if (!lSilentOperation)
                        printf("Filtering messages for player %d\n",i);
                    Filter();
                    if (AuxMessFile NEQ NULL)
                        SendMess();

                    ClearMessTree();
                    Write_Messages(i);
                    ClearData();
                }
        ClearAuxMess();
        chdir(lPreviousPath);
    }
    else
        printf("\n\ncan't access mess.def file\n");

    return 0;
}
