#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "conf.h"
#include "mess.h"
#include "messio.h"
#include "util.h"


/* This module does all necessary i/o operations for MESS. */



/*************************************************************************/
/*                        Miscellanous stuff                             */
/*************************************************************************/

#ifdef NEED_SYS_UNISTD_H
#include <sys/unistd.h>
#endif



char mResultName[20];


IO_Def InitFile(const char* pName, const char* pMode, FILE** pFile)
{
    assert((pName NEQ NULL) AND (pFile NEQ NULL));
    if ((*pFile = fopen(pName, pMode)) EQ NULL)
        return(IO_FAILURE);
    else
        return(IO_SUCCESS);
}



/* Byte conversion :
 * Convert long or short from little endian to big endian and vica versa
 */

#pragma warn -par
void ConvertLong(int32* pData, uns16 pNumber)
{
#ifdef DOS_CONVERT 
    while (pNumber-- > 0)
    {
        *pData =    ( ((*pData) & 0x000000FFUL) << 24 )
                  | ( ((*pData) & 0x0000FF00UL) <<  8 )
                  | ( ((*pData) & 0x00FF0000UL) >>  8 )
                  | ( ((*pData) & 0xFF000000UL) >> 24 );
        pData++;
    }
#endif
}

void ConvertShort(int16* pData, uns16 pNumber)
{
#ifdef DOS_CONVERT
    while (pNumber-- > 0)
    {
        *pData= ( (((*pData) & 0x00FFU) << 8) | (((*pData) & 0xFF00U) >> 8) );
        pData++;
    }
#endif
}
#pragma warn .par


/* own strtok version */
char* StrTok(char* Buffer)
{
    static char *lStrPtr,
                *lNextPtr;
    static char lStrBuffer[4000];


    if (Buffer NEQ NULL)
    {
        strcpy(lStrBuffer,Buffer);
        lStrPtr=lStrBuffer;
        lNextPtr=strchr(lStrPtr,'\r');
        if (lNextPtr NEQ NULL)
        {
            lNextPtr[0]=0;
            lNextPtr++;
        }
    }
    else
    {
        lStrPtr=lNextPtr;
        if (lNextPtr NEQ NULL)
        {
            lNextPtr=strchr(lStrPtr,'\r');

            if (lNextPtr NEQ NULL)
            {
                lNextPtr[0]=0;
                lNextPtr++;
            }
        }
    }
    return(lStrPtr);
}


void ToUpper(char* s)
{
    int i;
    for (i=0; i<strlen(s); i++)
        s[i]=toupper(s[i]);
}





/*************************************************************************/
/*                 Read/Write Messages from Result files                 */
/*************************************************************************/

/* Read_Messages : - Open file
 *                 - Seek to start of messages
 *                 - Read number of messages
 *                 - Read all messages
 */
IO_Def Read_Messages(int16 pPlayer)
{
    int32    lMessStart;
    FILE     *lResult;
    char     *lMessBuffer=(char*)MemAlloc(4000),
             *lMessLines[25];
    int16    i,j,
             LineCount;



    sprintf(mResultName,"player%d.rst",pPlayer);
    if (IO_SUCCESS NEQ InitFile(mResultName,"rb",&lResult))
    {
	sprintf(mResultName,"PLAYER%d.RST",pPlayer);
	if (IO_SUCCESS NEQ InitFile(mResultName,"rb",&lResult))
	    return(IO_FAILURE);
    }
    fseek(lResult,16,SEEK_SET);
    if (1-fread(&lMessStart,4,1,lResult))
        goto End;
    ConvertLong(&lMessStart,1);
    if (fseek(lResult,lMessStart-1,SEEK_SET))
        goto End;


    if (1-fread(&gMessageNr,2,1,lResult))
        goto End;
    ConvertShort(&gMessageNr,1);

    gMessages=(Message_Struct*)MemAlloc(sizeof(Message_Struct)*(gMessageNr+gAuxMessNr));
    memset(gMessages,0,sizeof(Message_Struct)*(gMessageNr+gAuxMessNr));

    for (i=0; i<gMessageNr; i++)
    {
        if (1-fread(&lMessStart,4,1,lResult))
            goto End;
        ConvertLong(&lMessStart,1);
        if (1-fread(&gMessages[i].Length,2,1,lResult))
            goto End;
        ConvertShort(&gMessages[i].Length,1);
    }

    fseek(lResult,1,SEEK_CUR);

    for (i=0; i<gMessageNr; i++) if (gMessages[i].Length NEQ 0)
    {
        if (1-fread(lMessBuffer,gMessages[i].Length,1,lResult))
            goto End;
	gMessages[i].Length=min(gMessages[i].Length,1000);
        for (j=0; j<gMessages[i].Length; j++)
        {
            if (lMessBuffer[j]==13)
                lMessBuffer[j]=45;
            lMessBuffer[j]-=13;
        }

        lMessBuffer[gMessages[i].Length]=0;


        LineCount=1;
        lMessLines[0]=StrTok(lMessBuffer);
        while (   ((lMessLines[LineCount]=StrTok(NULL)) NEQ NULL)
                &&(LineCount<25))
            LineCount++;

/* StrTok separates the lines by testing for \r. However, the last line could
   be \r\n or \n\r. The first would lResult in an extra line which we don't
   want. */
        if (   ((strlen(lMessLines[LineCount-1])==1) AND (lMessLines[LineCount-1][0]=='\n'))
             || (strlen(lMessLines[LineCount-1])==0)
           )
            LineCount--;

        gMessages[i].LineNr=LineCount;
        gMessages[i].Lines=(char**)MemAlloc(LineCount*sizeof(char*));

        for (j=0; j<LineCount; j++)
        {
            gMessages[i].Lines[j]=(char*)MemAlloc(strlen(lMessLines[j])+1);
            strcpy(gMessages[i].Lines[j],lMessLines[j]);
        }
    }

    fclose(lResult);

    return(IO_SUCCESS);

End: printf("\nunexpected end of file in %s\n",mResultName);
     fclose(lResult);
     return(IO_FAILURE);
}



/* Write_Messages : - Open result file
 *                  - Check if it is a v3.2 style result or not
 *                  - Change all header entries that need to be changed
 *                  - Seek to start of messages
 *                  - Write number of messages
 *                  - Write new messages
 *                  - copy the rest of the result
 */

IO_Def Write_Messages(int16 pPlayer)
{
    int32    lMessStart,
             lMessLength=3,
             lOldMessLength,
             lOffset,
             v35Start;
    FILE     *lResult,
             *lNewResult;
    char     lOldName[20],
             *lNewLine=(char*)MemAlloc(1001),
             *lMessBuffer;
    int16    i,j;


/* init names, remove .bak files, rename old results to .bak,
 * open necessary files */
    lMessBuffer=(char*)MemAlloc(4096);
    sprintf(lOldName,"PLAYER%d.BAK",pPlayer);
    remove(lOldName);
    sprintf(lOldName,"player%d.bak",pPlayer);
    remove(lOldName);

    if (rename(mResultName,lOldName))
    {
         printf("\ncan't find %s or file is read only\n",mResultName);
         exit(1);
    }
    if (IO_SUCCESS NEQ InitFile(lOldName,"rb",&lResult))
        return(IO_FAILURE);
    if (IO_SUCCESS NEQ InitFile(mResultName,"wb",&lNewResult))
        return(IO_FAILURE);

/* All data before the messages remains unchanged. Just copy the offsets */
    for (i=0; i<5; i++)
    {
        fread(&lMessStart,4,1,lResult);
        fwrite(&lMessStart,4,1,lNewResult);
    }

/* Now we have the message start. The difference between this value and the
 * next header entry tells us the previous length of the message section.
 * We sum the length of all messages and add 6 bytes for each message (for the
 * message header) to get the new length of the message section.
 */
    ConvertLong(&lMessStart,1);
    fread(&lOffset,4,1,lResult);
    ConvertLong(&lOffset,1);

    for (i=0; i<gMessageNr; i++)
    {
        for (j=0, gMessages[i].Length=gMessages[i].LineNr; j<gMessages[i].LineNr; j++)
            gMessages[i].Length+=strlen(gMessages[i].Lines[j]);
        lMessLength+=(gMessages[i].Length+6);
    }
    lOldMessLength=lOffset-lMessStart;
    lOffset=lMessStart+lMessLength;
    ConvertLong(&lOffset,1);
    fwrite(&lOffset,4,1,lNewResult);
    ConvertLong(&lOffset,1);

/* Adjust the remaining header entries. */
    for (i=0; i<2; i++)
    {
        fread(&lOffset,4,1,lResult);
        ConvertLong(&lOffset,1);
        lOffset+=(lMessLength-lOldMessLength);
        ConvertLong(&lOffset,1);
        fwrite(&lOffset,4,1,lNewResult);
        ConvertLong(&lOffset,1);
    }

/* Copy the next 8 bytes.
 * If they contain "Ver3.500" then we have a WinPlan style result and we need
 * to adjust the offset following the "Ver3.500" string, otherwise we have a
 * v3.0 style result.
 */
    fread(lMessBuffer,8,1,lResult);
    fwrite(lMessBuffer,8,1,lNewResult);
    fread(&v35Start,4,1,lResult);
    if (!strncmp(lMessBuffer,"Ver3.500",8))
    {
        ConvertLong(&v35Start,1);
        v35Start+=(lMessLength-lOldMessLength);
        ConvertLong(&v35Start,1);
    }
    fwrite(&v35Start,4,1,lNewResult);

/* Copy the result up to the message start. */
    for (i=0; i<(lMessStart-45)/4096; i++)
    {
        fread(lMessBuffer,4096,1,lResult);
        fwrite(lMessBuffer,4096,1,lNewResult);
    }
    fread(lMessBuffer,(int)((lMessStart-45)%4096),1,lResult);
    fwrite(lMessBuffer,(int)((lMessStart-45)%4096),1,lNewResult);


/* Write message section header. The header ends with a 0. */
    ConvertShort(&gMessageNr,1);
    fwrite(&gMessageNr,2,1,lNewResult);
    ConvertShort(&gMessageNr,1);
    lOffset=lMessStart+6*gMessageNr+3;
    for (i=0; i<gMessageNr; i++)
    {
        ConvertLong(&lOffset,1);
        fwrite(&lOffset,4,1,lNewResult);
        ConvertLong(&lOffset,1);
        ConvertShort(&gMessages[i].Length,1);
        fwrite(&gMessages[i].Length,2,1,lNewResult);
        ConvertShort(&gMessages[i].Length,1);
        lOffset+=gMessages[i].Length;
    }
    i=0;
    fwrite(&i,1,1,lNewResult);

/* Now write the new messages. */
    for (i=0; i<gMessageNr; i++)
    {
        lMessBuffer[0]=0;
        for (j=0; j<gMessages[i].LineNr; j++)
        {
            sprintf(lNewLine,"%s\r",gMessages[i].Lines[j]);
            strcat(lMessBuffer,lNewLine);
        }
        for (j=0; j<gMessages[i].Length; j++)
            lMessBuffer[j]+=13;
        fwrite(lMessBuffer,j,1,lNewResult);
    }

/* Copy the rest of the result. */
    fseek(lResult,lOldMessLength,SEEK_CUR);
    do
    {
        i=fread(lMessBuffer,1,4096,lResult);
        fwrite(lMessBuffer,1,i,lNewResult);
    }
    while (i);

    fclose(lResult);
    fclose(lNewResult);
    free(lNewLine);

    return(IO_SUCCESS);
}






/*************************************************************************/
/*                    Read message definition file                       */
/*************************************************************************/



/* Find [ID] sequences in a message definition, return number of [ID] lines */
int16 FindDef(DefLine_Struct* DefLines, char** DefBuffer, const char* IDStr,
              int16 LineCount)
{
    char    *Line, *s1, *s2, *s3, *s4;
    int16   Count=0, i;

    for (i=0; (i<LineCount); i++)
        if (strstr(DefBuffer[i],IDStr))
        {
            Line=strtok(DefBuffer[i],";");   /* cut comments */
            s3=strtok(Line,"\"");
            s2=strtok(NULL,"\"");            /* ident string */

            s1=strtok(s3," ");               /* should be IDStr */

            if ((s4=strtok(NULL," ")) NEQ NULL)
                DefLines[Count].Line=atoi(s4)-1;
            else
                DefLines[Count].Line=-1;

            if ((s4=strtok(NULL," ")) NEQ NULL)
                DefLines[Count].Word=atoi(s4);
            else
                DefLines[Count].Word=0;

            if ( (s1 EQ NULL) OR (strcmp(strtok(s1," "),IDStr))
                              OR (DefLines[Count].Line<0) )
            {
                printf("\nSyntax error in definition file. Program aborted.");
                exit(1);
            }
            else
            {
                if (s2!=NULL)
                {
                    DefLines[Count].Text=(char*)MemAlloc(strlen(s2)+1);
                    strcpy(DefLines[Count].Text,s2);
                }
                else
                    DefLines[Count].Text=NULL;
                Count++;
            }
        }
    return(Count);
}



/* The following function takes a section of a message definition file
 * that contains a definition (at least if the file is syntactically
 * correct) and parses for [ID], [NID], [IGNORE], [PRIORITY], [COMPARE],
 * [SUMMARY] and [HEADER] format lines.
 */

void FillDef(MessDef_Struct *pDef, FILE* pInp, char* pBuffer)
{
    int16            i,LineCount=0,
                     Count,
                     CompareLines[25];
    DefLine_Struct   DefLines[25];
    char             *DefBuffer[50],
                     *Line,
                     *s2;
    boolean          NewMess=false;


    memset(DefBuffer,0,sizeof(DefBuffer));

/* read whole message definition */
    do
    {
        fgets(pBuffer,255,pInp);
        for (i=0; i<strcspn(pBuffer,"]"); i++)
            pBuffer[i]=toupper(pBuffer[i]);
        for (i=0; i<4; i++)
            if (strstr(pBuffer,MessDefStrings[i]))
                NewMess=true;

        DefBuffer[LineCount]=(char*)MemAlloc(strlen(pBuffer)+1);
        strcpy(DefBuffer[LineCount],pBuffer);
        LineCount++;
    }
    while ( (!NewMess) AND (!feof(pInp)) AND (LineCount<50) );

/* look for message ID strings */

    if ((pDef->IdentNr=FindDef(DefLines,DefBuffer,IdentStr,LineCount)) EQ 0)
    {
        printf("\nmissing [ID] line in message definition, program abort\n");
        exit(1);
    }
    pDef->Ident=(DefLine_Struct*)MemAlloc(pDef->IdentNr*sizeof(DefLine_Struct));
    for (i=0; i<pDef->IdentNr; i++)
    {
        if (DefLines[i].Text EQ NULL)
        {
            printf("\nmissing definition string in [ID] line, program terminated\n");
            exit(1);
        }
        else
        {
            pDef->Ident[i]=DefLines[i];
            ToUpper(pDef->Ident[i].Text);
        }
    }

/* look for message NID strings */

    if ((pDef->NonIdentNr=FindDef(DefLines,DefBuffer,NIdentStr,LineCount)) NEQ 0)
    {
        pDef->NonIdent=(DefLine_Struct*)MemAlloc(pDef->NonIdentNr*sizeof(DefLine_Struct));
        for (i=0; i<pDef->NonIdentNr; i++)
        {
            if (DefLines[i].Text EQ NULL)
            {
                printf("\nmissing definition string in [IDN] line, program terminated\n");
                exit(1);
            }
            else
            {
                pDef->NonIdent[i]=DefLines[i];
                ToUpper(pDef->NonIdent[i].Text);
            }
        }
    }

/* look for [PRIORITY] format line */
    for (i=0; i<LineCount; i++)
        if (strstr(DefBuffer[i],PriorityStr))
        {
            Line=strpbrk(DefBuffer[i],"1234567890-");

            if (Line NEQ NULL)
                pDef->Priority=atoi(Line);
            else
            {
                printf("\n[PRIORITY] message should have a numeric priority definition\n");
                exit(1);
            }
        }

    if (pDef->Type EQ DOUBLE)
    {
/* look for [COMPARE] format line */
        Count=0;
        for (i=0; i<LineCount; i++)
            if (strstr(DefBuffer[i],CompareStr))
            {
                Line=strtok(DefBuffer[i]," -0");   /* cut comments */

                while (Line NEQ NULL)
                {
                    CompareLines[Count]=atoi(Line)-1;
                    if (CompareLines[Count]>20)
                        CompareLines[Count]=-1;

                    if (CompareLines[Count]>=0)
                        Count++;
                    Line=strtok(NULL," -0");
                }
            }
        if (Count>0)
        {
            pDef->CompareNr=Count;
            pDef->Compare=(int16*)MemAlloc(Count*2);
            for (i=0; i<Count; i++)
                pDef->Compare[i]=CompareLines[i];
        }


/* look for message [IGNORE] strings */

        if ((pDef->IgnoreNr=FindDef(DefLines,DefBuffer,IgnoreStr,LineCount)) NEQ 0)
        {
            pDef->Ignore=(DefLine_Struct*)MemAlloc(pDef->IgnoreNr*sizeof(DefLine_Struct));
            for (i=0; i<pDef->IgnoreNr; i++)
                pDef->Ignore[i]=DefLines[i];
        }
    }


    if (pDef->Type EQ SUMMARY)
    {
    /* look for message [SUMMARY] strings */
        if ((pDef->SummaryNr=FindDef(DefLines,DefBuffer,SummaryStr,LineCount)) NEQ 0)
        {
            pDef->Summary=(DefLine_Struct*)MemAlloc(pDef->SummaryNr*sizeof(DefLine_Struct));
            for (i=0; i<pDef->SummaryNr; i++)
                pDef->Summary[i]=DefLines[i];
        }
        else
        {
            printf("\n[SUMMARY] message must have one or more [EXTRACT] line(s)\n");
            exit(1);
        }



    /* look for [HEADER] lines */
        Count=0;
        for (i=0; i<LineCount; i++)
            if (strstr(DefBuffer[i],HeaderStr))
            {
                Line=strtok(DefBuffer[i],";");   /* cut comments */
                s2=strtok(Line,"\"");
                s2=strtok(NULL,"\"");            /* header string */

                if (s2 NEQ NULL)
                {
                    DefLines[Count].Text=(char*)MemAlloc(strlen(s2)+1);
                    strcpy(DefLines[Count].Text,s2);
                    Count++;
                }
                else
                {
                    printf("\nfound [HEADER] line without header text\n");
                    exit(1);
                }
            }

        if ((pDef->HeaderNr=Count) EQ 0)
        {
            printf("\nfound [SUMMARY] message without [HEADER] line(s)\n");
            exit(1);
        }
        else
        {
            pDef->SummaryHeader=(char**)MemAlloc(Count*sizeof(char*));
            for (i=0; i<Count; i++)
                pDef->SummaryHeader[i]=DefLines[i].Text;
        }

    /* look for [SUMMARY] format line */
        for (i=0; ((i<LineCount)&& (pDef->Type NEQ 0)); i++)
            if (strstr(DefBuffer[i],SummaryLineStr))
            {
                Line=strtok(DefBuffer[i],";");   /* cut comments */
                s2=strtok(Line,"\"");
                s2=strtok(NULL,"\"");            /* header string */

                if (s2 NEQ NULL)
                {
                    pDef->SummaryFormat=(char*)MemAlloc(strlen(s2)+1);
                    strcpy(pDef->SummaryFormat,s2);
                }
                else
                {
                    printf("\n[SUMMARY] message should have a [LINE] definition\n");
                    exit(1);
                }
            }

    }
}




/* Read in the message definition file */

IO_Def  ReadMessDefs(void)
{
    FILE             *DefinitionFile;
    int16            i;
    MessDef_Stack    *First, *Current;
    char             Buffer[255];


/* Init our definition stuff */
    First=(MessDef_Stack*)MemAlloc(sizeof(MessDef_Stack));
    memset(First,0,sizeof(MessDef_Stack));
    First->Succ=NULL;
    Current=First;

    if (InitFile(MessFile,"rt",&DefinitionFile) NEQ IO_SUCCESS)
        return(IO_FAILURE);
    fgets(Buffer,255,DefinitionFile);

    do
    {
/* Read until we encounter the start of a definition */
        do
        {
            for (i=0; i<strcspn(Buffer,"]"); i++)
                Buffer[i]=toupper(Buffer[i]);
            for (i=0; i<4; i++)
                if (strstr(Buffer,MessDefStrings[i]))
                    Current->Data.Type=i+1;
            if (Current->Data.Type EQ 0)
                fgets(Buffer,255,DefinitionFile);
        }
        while ( (!feof(DefinitionFile)) AND (Current->Data.Type EQ 0) );

/* Read the definition */
        if (Current->Data.Type NEQ 0)
        {
            FillDef(&Current->Data,DefinitionFile,Buffer);
            DefCount++;
            Current->Succ=(MessDef_Stack*)MemAlloc(sizeof(MessDef_Stack));
            memset(Current->Succ,0,sizeof(MessDef_Stack));
            Current=Current->Succ;
            Current->Succ=NULL;
        }
    }
    while (!feof(DefinitionFile));

    gMessDefs=(MessDef_Struct*)MemCAlloc(DefCount,sizeof(MessDef_Struct));

    Current=First;
    for (i=0; i<DefCount; i++)
    {
        gMessDefs[i]=Current->Data;
        Current=Current->Succ;
        MemFree(First);
        First=Current;
    }
    MemFree(Current);

    return(IO_SUCCESS);
}
