/*
some constants
*/

#define RACE_NR            11
#define SHIP_NR           500
#define PLANET_NR         500
#define BASE_NR           500
#define MINE_NR           500
#define HULL_NR           105
#define ENGINE_NR           9
#define BEAM_NR            10
#define TORP_NR            10
#define RACEHULLS          20

#define MAX_SPEED           9
#define MAX_BEAMS          10
#define MAX_TUBES          10
#define MAX_BAYS           10
#define MAX_BASE_DEFENSE  200
#define MAX_BASE_FIGHTERS  60
#define MAX_TECH           10
#define MAX_DENSITY       100
#define MAX_HAPPY         100
#define MAX_COORDINATE  10000   /* coordinate limits given by planets.exe */
#define MIN_COORDINATE      0

#define PLANET_SIZE        85
#define BASE_SIZE         156
#define SHIP_SIZE         107
#define MINE_SIZE          16
#define HULLSPEC_SIZE      60
#define ENGSPEC_SIZE       66
#define TORPSPEC_SIZE      38
#define BEAMSPEC_SIZE      36
#define TRUEHULL_SIZE     440
#define PLANETNAME_SIZE    20
#define VCR_SIZE          100
#define HCONFIG_SIZE      302
#define RACENAME_SIZE     682
#define LONGNAME_SIZE      30
#define SHORTNAME_SIZE     20
#define ADJECTIVE_SIZE     12
#define SHIPXY_SIZE         8

typedef enum {NEUTRONIUM=0, TRITANIUM=1, DURANIUM=2, MOLYBDENIUM=3,
              COLONISTS=4, SUPPLIES=5, CREDITS=6}
CargoType_Def;


typedef enum {NOMEM}
ErrorType_Def;


typedef enum {NoOwner=0, Federation=1, Gorn=2, Romulans=3, Klingons=4, Orions=5,
              Borg=6, Tholians=7, Empire=8, Cylons=9, Rebels=10, Colonies=11}
RaceType_Def;


typedef enum {NoNatives=0, Humanoid=1, Bovinoid=2, Reptilian=3, Avian=4,
              Amorphous=5, Insectoid=6, Amphibian=7, Ghipsoldal=8,
              Siliconoid=9}
NativeType_Def;


typedef enum {NoGovm=0, Anarchy=1, PreTribal=2, EarlyTribal=3, Tribal=4,
              Feudal=5, Monarchy=6, Representative=7, Participatory=8,
              Unity=9}
NativeGovm_Def;


typedef enum {NoMission=0, Exploration=1, MineSweep=2, LayMines=3, Kill=4,
              SensorSweep=5, Colonize=6, Tow=7, Intercept=8, SpecialMission=9,
              Cloak=10, GatherNeutronium=11, GatherTritanium=12,
              GatherDuranium=13, GatherMolybdenium=14, GatherSupplies=15}
ShipMission_Def;


typedef enum {NoBaseMiss=0, Refuel=1, RepairBase=2, MaxDefense=3,
              UnloadFreighters=4, LoadTorps=5, ForceSurrender=6}
BaseMission_Def;

typedef enum {FIX=0, RECYCLE=1}
BaseFixMission_Def;


typedef enum {HULL=0, ENGINE=1, BEAM=2, TORP=3}
BaseTech_Def;



/*************************************************************************
 host data
 *************************************************************************/

typedef struct
{
    RaceType_Def   Owner;
    int16          Id;
    char           FriendlyCode[3];
    int16          Mines,
                   Factories,
                   Defense;
    int32          Cargo[7],      /* N, T, D, M, C, S, MC */
                   Core[4];
    int16          Density[4],
                   ColTax,
                   NatTax,
                   ColHappy,
                   NatHappy;
    NativeGovm_Def Govm;
    int32          NativePopulation;
    NativeType_Def NativeType;
    int16          Temp;
    boolean        BuildBase;
}
Planet_Struct;


/* pdata.hst : int16 <unknown>, Planet_Struct[500] planetdata */



typedef struct
{
    int16           Id;
    RaceType_Def    Owner;
    char            FriendlyCode[3];
    int16           Speed,
                    Waypoint[2],
                    Location[2],          /* x,y */
                    Engine,
                    Hull,
                    BeamType,
                    BeamNumber,
                    Bays,
                    TorpType,
                    Ammunition,
                    TubeNumber;
    ShipMission_Def Mission;
    RaceType_Def    Enemy;               /* primary enemy */
    int16           TowTarget,
                    Damage,
                    Crew,
                    Colonists;
    char            Name[20];            /* NO closing 0 !!! */
    int16           Cargo[5],            /* N, T, D, M, S */
                    Dump[6],             /* N, T, D, M, C, S */
                    DumpPlanet,          /* 0 for jettison */
                    Transfer[6],         /* N, T, D, M, C, S */
                    TransferShip,
                    InterceptTarget,
                    Credits;
}
Ship_Struct;


/* ship.hst : int16 <unknown>, Ship_Struct shipdata[500] */



typedef struct
{
    int16             Id;
    RaceType_Def      Owner;
    int16             Defense,
                      Damage,
                      TechLevel[4],
                      Engines[9],
                      Hulls[20],
                      Beams[10],
                      Tubes[10],
                      Torps[10],
                      Fighters,
                      FixingShip;
   BaseFixMission_Def FixOrder;
   BaseMission_Def    Order;
   int16              ToBuild[7];  /* hull,eng,beam,#beams,torp,#torp,bla */
}
Base_Struct;


/* bdata.hst : int16 <unknown>, vgabase basedata[500], int16 turnnumber */




typedef struct
{
    int16        X,Y;
    RaceType_Def Owner;
}
XYData_Struct;

/* xyplan.hst, xyplan.dat : Xydata_Struct[500] ; xyplan.dat : owner always 0 */






typedef struct
{
    char     Name[30];
    int16    Picnumber,
             Bla,              /* always 1 ? */
             TritCost,
             DurCost,
             MolyCost,
             FuelCapacity,
             CrewComplement,
             EngineNumber,
             Mass,
             TechLevel,
             CargoCapacity,
             Bays,
             TubeNumber,
             BeamNumber,
             MoneyCost;
}
Hullspec_Struct;


typedef struct
{
    char     Name[20];
    int16    MoneyCost,
             TritCost,
             DurCost,
             MolyCost,
             TechLevel;
    int32    FuelConsumption[9];
}
Engspec_Struct;



typedef struct
{
    char     Name[20];
    int16    TorpCost,
             TubeCost,
             TubeTritCost,
             TubeDurCost,
             TubeMolyCost,
             TubeMass,
             TechLevel,
             KillPower,
             DestructivePower;
}
Torpspec_Struct;



typedef struct
{
    char     Name[20];
    int16    MoneyCost,
             TritCost,
             DurCost,
             MolyCost,
             Mass,
             TechLevel,
             KillPower,
             DestructivePower;
}
Beamspec_Struct;

/* hullspec.dat : Hullspec_Struct[105]
   engspec.dat  : Engspec_Struct[9]   + byte unused[4]
   beamspec.dat : Beamspec_Struct[10]
   torpspec.dat : Torpspe_Struct[10]  + byte unused[90] */





typedef int16       Truehull_Def[12][21];

/* Every race can build up to 20 different hulls
   the numbers in truehull.dat refer to the hull slot in hullspec.dat
   ID's range from 1 to 105 (=HULLNR)
   truehull contains zeros if the race can build less than 20 hulls
*/





typedef struct
{
    int16        X,
                 Y;
    RaceType_Def Race;
    int32        Units;
    boolean      Web;
}
MineField_Struct;

/* mines.hst : MineField_Struct[500] */




typedef char  *RaceName_Def[3][12];



typedef struct
{
    int16        X,
                 Y;
    RaceType_Def Owner;
    int16        Mass;
}
Shipxy_Struct;

/* shipxy.hst : shipxy[500] */



typedef struct
{
    char   HostTime[18];
    int16  TurnNumber;
}
Turntime_Struct;

/* nextturn.hst, lastturn.hst, both identical ? */



typedef struct
{
    int16    RecycleRate,
             MeteorRate;
    boolean  Minefields,
             Alchemy,
             DeleteOldMessages,
             DisablePasswd;
    int16    GroundKill[12],        /* tim uses only 1..11, 0 ignored */
             GroundDefense[12],
             FreeFighters[12],
             MiningRate[12],
             TaxRate[12];
    boolean  RebelsBuildFighters,
             ColoniesBuildFighters,
             RobotsBuildFighters;
    int16    CloakFailure;
    boolean  RobCloakedShips;
    int16    ScanRange,
             DarksenseRange;
    boolean  Hiss,
             RebelGroundAttack,
             SuperRefit,
             WebMines;
    int16    CloakFuelBurn,
             SensorRange;
    boolean  NewNatives,
             Nuk;
    int16    AssimilationRate,
             WebDecay,
             MineDecay,
             MaxSize,
             TransuraniumRate,
             StructureDecay;
    boolean  EatSupplies,
             NoFuelMove;
    int16    MineOdds,
             WebOdds,
             MineScanRange;
    boolean  MinesDestroyMines,
             EngShieldBonus;
    int16    ShieldBonusRate,
             ColFighterSweepRate;
    boolean  ColSweepWebs;
    int16    MineSweepRate,
             WebSweepRate,
             HissEffect,
             RobFailRate;
    boolean  RebelNoAtt,
             FascNoAtt;
    int16    MineDist,
             WebDist;
    boolean  ScienceShips;
    int16    CloakMineOdds,         /*  * 0.1 */
             NoCloakDamage;
    boolean  FedCrewBonus;

    int16    SmallMeteorOdds;
    int32    SmallOre[9];           /* 4 min, 4 max, Smallore[0] ignored */
    int16    LargeMeteorNumber;
    int32    LargeOre[9];
    boolean  MeteorMessages;

    boolean  OneTow,
             HyperDrives;
    int16    ClimateDeathRate;
    boolean  GravityWells,
             CrystalsLikeDesert;
    boolean  DestroyWebs,
             ClimateLimitsPopulation;
}
Hconfig_Struct;

/* hconfig.hst, oldfig.hst : Hconfig_Struct */






typedef struct
{
    char         Name[20];
    int16        Damage,
                 Crew,
                 Id;
    RaceType_Def Race;
    int16        Picnumber,    /* picture in hullspec (picnumber) */
                 Beamtype,
                 Beamnumber,
                 Bays,
                 Torps,
                 Ammunition,   /* torps or fighters */
                 Tubes;
}
ShipVcr_Struct;



typedef struct
{
    int16            Random1;
    boolean          Dummy;               /* never true */
    int16            Planetpic;           /* only !=0 if is_planet2 is true */
    boolean          Is_planet2;
    int16            Mass1,
                     Mass2;
    ShipVcr_Struct   Ship1,
                     Ship2;
    int16            Shield1,
                     Shield2;
}
Vcr_Struct;


/*
structure of vcr??.dat/vcr.hst :

int16        number_of_battles;
Vcr_Struct   vcrdat[number_of_battles];
byte         signature[10];  (player data only)
*/



/* in gen.hst are a lot bytes with unknown use, which our host
   can use for his own infos */

typedef struct
{
    byte     Bla[46];
    boolean  Players[RACE_NR];           /* active players, identical to init.tmp */
    char     Passwd1[RACE_NR][10],
             Passwd2[RACE_NR][10];
/* this is all fromm address 0000-011F after master (= 288 Bytes)
   after first hosting it will be 450 Bytes
       +byte     bla2[162];
   mostly bla and bla2 have only 00
*/
}
HostGen_Struct;




/*************************************************************************
 result file
 *************************************************************************

int32 header[8];      ( contains starting adress of each components )
const byte  bla=0;

bodies file1...file8 :

1 ship??.dat
2 target??.dat
3 pdata??.dat
4 bdata??.dat
5 mdata??.dat
6 shipxy??.dat
7 gen??.dat
8 vcr??.dat


    * all bodies are WITHOUT the last 10-byte-playerlog
      (messages never have it)
    * in gen??.dat pw1 is ok and pw2 = playerlog_for_dis-files
    Now you can calculate the dat-playerlog from this dis-playerlog
    dat-playerlog[i] = ( dis-playerlog[i] + i )

The planet structure remains the same, pdata.dat, bdata.dat and ship.dat
start with an int16 for the number of entries and continue with the entries
ordered by ID.

shipxy.dat has always 500 entries.


mdata.dat starts with an int16 (number of messages) followed by a
header for each message which consists of the starting adress (int32)
and the length. After the header follows one 0 and then the bodies.

The adresses in the result are relative to the start of the result
(although I don't understand why Tim doesn't make them relative to
the start of mdata.dat, unpack converts them).




/*************************************************************************
 turn file
 *************************************************************************/


typedef struct
{
    int16     Owner;
    int32     Commands;       /* int32 ??? >32000 commands ? :-) */
    char      Hosttime[18];
    int16     Bla1,
              Hostlog_Chk;    /* byte-checksum over hosttime */
}
Header_Struct;


/*

after the header :
if no commands -> turn_end
if command# >0 -> 1 dummy byte
                  int32[commands]   adresses of commands relative to the turn file start
                  commands

*/

typedef struct
{
    int32        Race_Chk,
                 Chksum;
    int32        Reg_info[2][25];
    int32        Reg_checksum;
    int32        Race_checksum[RACE_NR];
}
TurnEnd_Struct;


/* register log starts at filesize-$f8

commands :
common structure is <int16 command> <int16 id> <parameters>
Exceptions : message,password

List of commands & parameters

comm object parameter meaning

01   ship   char[3]     fcode
02   ship   int16       warp
03   ship   int16       way (x,y)
04   ship   int16       mission
05   ship   int16       primary_enemy
06   ship   int16       to_tow
07   ship   char[20]    name
08   ship   int16[7]    dump: fuel,trit,dur,moly,col,suppl,planet (=00 for jettison)
09   ship   int16[7]    transfer (fuel,trit,dur,moly,col,suppl,ship_to_transfer)
0a   ship   int16       to_intercept
0b   ship   int16       fuel
0c   ship   int16       trit
0d   ship   int16       dur
0e   ship   int16       moly
0f   ship   int16       suppl
10   ship   int16       col
11   ship   int16       fighter/torps
12   ship   int16       MC
13   -
14   -
15   plan   char[3]     fc
16   plan   int16       mines
17   plan   int16       fact
18   plan   int16       defense
19   plan   int32       fuel
1a   plan   int32       trit
1b   plan   int32       dur
1c   plan   int32       moly
1d   plan   int32       col
1e   plan   int32       suppl
1f   plan   int32       MC
20   plan   int16       tax_col
21   plan   int16       tax_nat
22   plan               build base (no parameter)
23   -
24   -
25   -
26   -
27   -
28   base   int16       defense
29   base   int16       eng_level
2a   base   int16       hull_level
2b   base   int16       beam_lev
2c   base   int16[9]    engs
2d   base   int16[20]   hulls
2e   base   int16[10]   beams
2f   base   int16[10]   tubes
30   base   int16[10]   torps
31   base   int16       fighters
32   base   int16       ship_to_fix/recycle
33   base   int16       order fix(1)/recycle(2)
34   base   int16       primary_order
35   base   int16[7]    build ship (base-ID,hull,engT,beamT,beam#,torpT,torp#,bla)
36   base   int16       torp_level
37   -
38   -
39   -
3a   -
3b   -
3c                      send message (length,from,to,body)
3d                      new_password 3d 00 (+ 10 byte for password, shifted + 0x32)

*/




